#ifndef EXPLICITEULERSCHEME_HH
#define EXPLICITEULERSCHEME_HH

#include <cassert>
#include <functional>

template<typename TimeType, typename Domain, typename Range>
using TimeFunction = std::function<Range(const TimeType&, const Domain&)>;

/**
 * Interface for one step time integration methods which are usable to
 * numerically solve the initial value problem
 *   y' = f(t,y) on [t_0,T], y(t_0) = y_0,
 * with f: [t_0,T] x R^n --> R^n, t_0,T \in R_0^+, y_0 \in R^n, n \in N.
 * Note that this interface might be too restrictive in general. Its purpose
 * is to give you an idea about how such an interface could look like.
 * @tparam VectorType The type representing a vector in R^n; it is expected
 *                    to provide an operator[] method for element access and
 *                    a size method like known from STL containers.
 * @tparam TimeType   Type to represent time values.
 */
template <typename VectorType, typename TimeType = double>
class OneStepTimeIntegrator
{
  /**
   * Do one step of the one step time integration method.
   * @param[in]     t     Start of time step.
   * @param[in]     dt    Time step size.
   * @param[in]     y_old Value at the beginning of time step.
   * @param[in,out] y_new Value at the end of time step; contains
   *                      initial guess e.g. for implicit methods.
   */
  virtual void apply (const TimeType& t, const TimeType& dt,
                      const VectorType& y_old, VectorType& y_new) const = 0;
};


/**
 * The explicit Euler time-stepping scheme for numerically solving the
 * initial value problem
 *   y' = f(t,y) on [t_0,T], y(t_0) = y_0,
 * with f: [t_0,T] x R^n --> R^n, t_0,T \in R_0^+, y_0 \in R^n, n \in N.
 * @tparam VectorType The type representing a vector in R^n; it is expected
 *                    to provide an operator[] method for element access and
 *                    a size method like known from STL containers.
 * @tparam TimeType   Type to represent time values.
 */
template <typename VectorType, typename TimeType = double>
class ExplicitEulerScheme
  : public OneStepTimeIntegrator<VectorType,TimeType>
{
public:
  using RhsFunction = TimeFunction<TimeType,VectorType,VectorType>;

  ExplicitEulerScheme (const RhsFunction f)
    : f_(f)
  {}

  /**
   * Do one step of the explicit Euler time-stepping scheme.
   * @param[in]  t     Start of time step.
   * @param[in]  dt    Time step size.
   * @param[in]  y_old Value at the beginning of time step.
   * @param[out] y_new Value at the end of time step.
   */
  virtual void apply (const TimeType& t, const TimeType& dt,
                      const VectorType& y_old, VectorType& y_new) const
  {
    // your code
  }

private:
  const RhsFunction f_;
};


#endif // EXPLICITEULERSCHEME_HH
